'
' Demonstrate a 32 element fixed point FFT routine used as a colour organ
' Each frequency range is sent to one of eight LEDs, 
' whose brightness is adjusted according to the energy seen at that particular frequency band
' 
' The code also implements a Special event interrupt that constantly takes ADC samples
' and a Timer0 interrupt that implements a software PWM of the LEDs
'
' Written for the Crownhill Proton compiler version 3.5.3.0 onwards using an 18F device
'
' Author Les Johnson 02-02-2012 
'
' You are free to use this code as you see fit, however, a mention of the author and compiler would be appreciated.
' The code is offered as-is and Crownhill nor the author take any responsibility for its use.
' 
    Include "Amicus18.inc"                  ' Configure for the Amicus18 device (18F25K20) running at 64MHz
    Include "Amicus18_ADC.inc"              ' Load the Amicus18 ADC routines into the program
    Include "Amicus18_Timers.inc"           ' Load the Amicus18 Timer macros into the program
    
    On_Hardware_Interrupt GoTo ISR_Handler  ' Point interrupt to handler
    
    $define FFT_Use_Sqr                     ' Instruct the FFT routine to use more refined real and imaginary combining
    $define FFT_Elements 32                 ' The amount of elements for the FFT routine to work on
    Include "FFT.inc"                       ' Load the 8-bit FFT routines into the program
'
' Setup the Alphanumeric LCD
'    
    Declare LCD_DTPin = PORTB.4             ' LCD's Data lines (D4 to D7)
    Declare LCD_ENPin = PORTB.3             ' LCD's EN line
    Declare LCD_RSPin = PORTB.2             ' LCD's RS line
    Declare LCD_Interface = 4               ' 4-bit interface to LCD
    Declare LCD_Lines = 2                   ' LCD contains 2 lines
    Declare LCD_Type = Alphanumeric         ' LCD type is alphanumeric    
'    
' Create Variables used within the software PWM interrupt
'
    Dim PWM_bPortShadow As Byte System
    Dim PWM_bAccum[8] As Byte               ' PWM_Duty Cycle Accumumulators  
    Dim PWM_bDutyCycle[8] As Byte           ' Duty Cycle registers
    Dim bDuty As Byte
'
' Create variables for the sampling interrupt
'
    Dim bSampleFlags As Byte System 
    Dim tSampleReady As bSampleFlags.0
    Dim bSampleIndex As Byte System
    Dim SampleArray[cFFT_NumberOfSamples] As Byte Heap
    Dim wCCPR2 As CCPR2L.Word               ' Combine CCP2L\H into a 16-bit register
    Dim wFSR0 As FSR0L.Word                 ' Combine FSR0L\H into a 16-Bit register
'
' Create variables used within the Map8 subroutine
'
    Dim Map_bTemp1 As Byte
    Dim Map_bTemp2 As Byte
    Dim Map_wTemp As Word                   ' Holds the result of the scale calculation   
    Dim Map_bXin As Map_bTemp1              ' Value is the value to Scale
    Dim Map_bInMin As Byte                  ' Lower bound of the value's current range
    Dim Map_bInMax As Byte                  ' Upper bound of the value's current range
    Dim Map_bOutMin As Byte                 ' Lower bound of the value's target range
    Dim Map_bOutMax As Map_bTemp2           ' Upper bound of the value's target range 
'
' Create variables used for display of the FFT spectrum
'    
    Dim wHighestOfTwo As Word               ' Holds two elements extracted from the FFT real array
    Dim bIndex As Byte                      ' General purpose index variable
    Dim bLED_Pos As Byte                    ' Holds the x position on the LCD
    Dim wCharData As Word                   ' Holds the two individual LCD bar patterns
    Dim bEnergyValue As Byte                ' Holds the value for the bar display on the LCD 
    Dim wPreviousValue As SWord             ' Holds the previous duty cycle value
    
$ifndef True
    $define True 1
$endif  
$ifndef False
    $define False 0
$endif    
'--------------------------------------------------------------------------------    
    GoTo Main                               ' Jump over the subroutines                                     
'------------------------------------------------------------
' Interrupt Subroutine
' Capture ADC samples and place them in an array
' Update eight 8-bit PWMs on PORTB
'
ISR_Handler:
    Context Save FSR0L, FSR0H
'
' Do a special event interrupt on CCP2 and Timer3
' This captures ADC samples and places them in an array
'    
    If PIR2bits_CCP2IF = 1 Then                             ' Was it a Compare Special Event on CCP2 that triggered the interrupt?
        Dec bSampleIndex                                    ' Yes. So. decrement the sample array index
        wFSR0 = VarPtr SampleArray                          ' Get the address of the sample array
        wFSR0 = wFSR0 + bSampleIndex                        ' Add the array index to it
        Repeat : Until ADCON0bits_GO_DONE = 0               ' Wait for the ADC to complete it's aquisition
        INDF0 = Abs ADRESH                                  ' Place the 8-bit ADC result into the sample array
        If bSampleIndex = 0 Then                            ' Have we reached the end of the samples array?
            bSampleIndex = SizeOf SampleArray               ' Yes. So reset the array index size
            tSampleReady = True                             ' Signal that the array has been filled with samples 
        EndIf
        PIR2bits_CCP2IF = 0                                 ' Clear the interrupt flag  
    EndIf
'
' Do a Timer0 overflow interrupt
' This implements the software PWM of eight LEDs
'    
    If INTCONbits_T0IF = 1 Then                             ' Was it a Timer0 overflow that triggered the interrupt?
        PWM_bAccum#0 = PWM_bAccum#0 + PWM_bDutyCycle#0      ' Yes. So Update Duty Cycle 0
        Rol PWM_bPortShadow    
        PWM_bAccum#1 = PWM_bAccum#1 + PWM_bDutyCycle#1      ' Update Duty Cycle 1
        Rol PWM_bPortShadow    
        PWM_bAccum#2 = PWM_bAccum#2 + PWM_bDutyCycle#2      ' Update Duty Cycle 2
        Rol PWM_bPortShadow    
        PWM_bAccum#3 = PWM_bAccum#3 + PWM_bDutyCycle#3      ' Update Duty Cycle 3
        Rol PWM_bPortShadow    
        PWM_bAccum#4 = PWM_bAccum#4 + PWM_bDutyCycle#4      ' Update Duty Cycle 4
        Rol PWM_bPortShadow    
        PWM_bAccum#5 = PWM_bAccum#5 + PWM_bDutyCycle#5      ' Update buty Cycle 5
        Rol PWM_bPortShadow    
        PWM_bAccum#6 = PWM_bAccum#6 + PWM_bDutyCycle#6      ' Update buty Cycle 6
        Rol PWM_bPortShadow    
        PWM_bAccum#7 = PWM_bAccum#7 + PWM_bDutyCycle#7      ' Update buty Cycle 7
        Rlf PWM_bPortShadow,W
        PORTB = WREG                                        ' Output Duty Cycle States
        INTCONbits_T0IF = 0                                 ' Clear the Timer0 overflow flag
    EndIf
    
    Context Restore                                         ' Exit the interrupt

'-----------------------------------------------------------------------------------------
' Scale one 8-bit value to another 8-bit value
' Input     : Map_bXin holds the value to scale
'           : Map_bInMin is the lower bound of the value's current range
'           : Map_bInMax is the upper bound of the value's current range
'           : Map_bOutMin is the lower bound of the value's target range
'           : Map_bOutMax is the upper bound of the value's target range
' Output    : pResult holds the result
' Notes     : Traps a zero value for extra speed
'  
$define Map8(pX, pInMin, pInMax, pOutMin, pOutMax, pResult) '
    Map_bXin = pX         '
    Map_bInMin = pInMin   '
    Map_bInMax = pInMax   '
    Map_bOutMin = pOutMin '
    Map_bOutMax = pOutMax '
    _Map8                 '
    pResult = Map_wTemp

_Map8 Macro-
    GoSub __Map8
Endm

#ifMacro- _Map8
__Map8:
    If Map_bXin = 0 Then                        ' Is the value to scale zero?
        Map_wTemp = 0                           ' Yes. So return a zero
    Else                                        ' Otherwise...
        Map_bTemp1 = Map_bXin - Map_bInMin
        Map_bTemp2 = Map_bOutMax - Map_bOutMin
        Map_wTemp = Map_bTemp1 * Map_bTemp2
        Map_bTemp1 = Map_bInMax - Map_bInMin
        Map_bTemp2 = Map_wTemp / Map_bTemp1
        Map_wTemp = Map_bTemp2 + Map_bOutMin
    EndIf
    Return  
#endIfMacro-
'------------------------------------------------------------
' Initialise the Software PWM interrupt
' Input     : None
' Output    : None
' Notes     : None
'
Initialise:
'
' Open the ADC for samples taken on AN0
' Left justify for 8-bit result
' External Vref+
'
    OpenADC(ADC_FOSC_64 & ADC_LEFT_JUST & ADC_2_TAD, ADC_REF_VREFPLUS_VSS, ADC_1ANA)
    
    CCP2CON = %00001011                             ' Compare mode: trigger special event, reset timer, start A/D conversion on CCP2 match (tCCP2IF bit is set)
'
' Setup Timer3 
'       Clear TMPR3L\H
'       Timer3 Interrupt disabled
'       Timer3 Width 16-bit read-write mode
'       Timer3 Internal clock source (TOSC)
'       Timer3 1:4 prescaler
'       Don't sync external clock input
'       Make Timer3 the source for a compare
'
    OpenTimer3(TIMER_INT_OFF & T3_16BIT_RW & T3_PS_1_4 & T3_SYNC_EXT_OFF & T3_SOURCE_INT & T3_SOURCE_CCP)
    $define cPrescalerValue 4                      ' Alter this to match the prescaler parameter above. i.e. 4 for T3_PS_1_4 
'
' Frequency calculations are:
' Time (in uS) = 1000000 / Frequency (in Hz)
' Frequency (in Hz) = 1000000 / Time (in uS)
'     
    $define cMicroSeconds $eval 1000000 / 8000     ' Interrupt rate (in uS) for an 8KHz sample rate and a 4KHz FFT spectrum
'
' Calculate the value to place into the CCPRx register in order to achieve a certain interrupt rate (in us)
'   
    $define cCCPR_Value $eval (cMicroSeconds / cPrescalerValue) * (_xtal / 4)

$if cCCPR_Value > 65535
    $error "Value too large for interrupt duration"
$elseif cCCPR_Value = 0
    $error "Value too small for interrupt duration"
$endif

    ReadADC(ADC_CH0)                                ' A dummy read of the ADC to setup the channel for the interrupt
    bSampleIndex = SizeOf SampleArray
    tSampleReady = False
    wCCPR2 = cCCPR_Value                            ' Load CCPR2L\H with the value to trigger an interrupt at a certain duration
    PIE2bits_CCP2IE = 1                             ' Enable the Special Event Interrupt on CCP2
        
    PWM_bPortShadow = 0
    Clear PWM_bAccum       
    Clear PWM_bDutyCycle   
    Low PORTB                               ' Make PortB an output and extinguish the LEDs   
'
' Configure Timer0 for:
'                       Clear TMR0L and TMR0H registers
'                       Interrupt on Timer0
'                       8-bit operation
'                       Internal clock source
'                       1:32 Prescaler
'
    OpenTimer0(TIMER_INT_ON & T0_8BIT & T0_SOURCE_INT & T0_PS_1_32)

    INTCONbits_PEIE = 1                             ' Enable Peripheral Interrupts
    INTCONbits_GIE = 1                              ' Enable Global Interrupts
    Return
'-------------------------------------------------------------------------
' Illuminate one of eight LEDs according to the frequency range and energy of an audio sample
' Input     : bEnergyValue holds the value to use for the PWM brightness of an LED
'           : bLED_Pos is the current LED position
' Output    : None
' Notes     : Implements a fade once the LED is not being illuminated
'
$define DisplayLEDs() GoSub _DisplayLED

_DisplayLED:
    If bEnergyValue > 3 Then                                    ' Is there an LED to illuminate?    
        If bEnergyValue >= 32 Then bEnergyValue = 32            ' Make sure a maximum is not breached
        Map8(bEnergyValue, 0,32, 0, 255, bDuty)                 ' Scale the FFT result into something more suitable as PWM
        PWM_bDutyCycle[bLED_Pos] = bDuty                        ' Adjust the current LEDs brightness 
    Else                                                        ' Otherwise... Fade the LED back to 0
        wPreviousValue = PWM_bDutyCycle[bLED_Pos]               ' Get the LEDs duty value
        If bLED_Pos > 1 Then                                    ' Are we accessing the first two LEDs. i.e. low frequencies
            wPreviousValue = wPreviousValue - 2                 ' No. So a slower fade
        Else                                                    ' Otherwise...
            wPreviousValue = wPreviousValue - 8                 ' A faster fade because these are bass sounds
        EndIf
        If wPreviousValue < 0 Then wPreviousValue = 0           ' Trap a value less than 0
        PWM_bDutyCycle[bLED_Pos] = wPreviousValue               ' Adjust the LEDs brightness
    EndIf
    Return 
'------------------------------------------------------------------------- 
' The main program loop starts here 
'
Main:
    GoSub Initialise
       
    While 1 = 1                                                 ' Create an infinite loop      
        If tSampleReady = True Then                             ' Is there a sample waiting
            Str FFT_bRealData = Str SampleArray                 ' Yes. So copy the samples array into the FFT_bRealData array
            tSampleReady = False                                ' Reset the flag
            Fix_FFT()                                           ' Perform the FFT (Result in array FFT_bRealData)
            '
            ' Display the spectrum on a 16x2 alphanumeric LCD
            ' As columns of vertical bars
            '    
            bLED_Pos = 0                                        ' Start at column 1 of the LCD
            bIndex = 0                                          ' Create a loop for all the values in the real array
            Repeat
                '
                ' We only have 8 LEDs, but 16 pieces of data from the FFT, So...
                ' Extract two values from the real array
                ' Find the maximum of the two values and use this as the duty cycle of an LED's PWM
                '
                wHighestOfTwo.Byte0 = FFT_bRealData[bIndex]         ' Extract the first value
                Inc bIndex
                wHighestOfTwo.Byte1 = FFT_bRealData[bIndex]         ' Extract the second value
                '
                ' Find the highest value of the two and load bEnergyValue with it
                '
                bEnergyValue = wHighestOfTwo.Byte0                  ' Default to bEnergyValue = the first value
                If bEnergyValue < wHighestOfTwo.Byte1 Then          ' Is bEnergyValue less than the second value?
                    bEnergyValue = wHighestOfTwo.Byte1              ' Yes. So bEnergyValue = the second value
                EndIf
                
                DisplayLEDs()                                       ' Illuminate the current LED
                Inc bLED_Pos                                        ' Next LED
                If bLED_Pos > 7 Then bLED_Pos = 0                   ' Back to LED 1 when the last LED is reached
                Inc bIndex
            Until bIndex >= cFFT_NumberOfRealElements               ' Until all the elements of the FFT are read
        EndIf
    Wend                                                            ' Do it forever
    
